#ifndef lint
static char SCCSid[] = "%W% %G%";
#endif

#include "mpid.h"

/* 
   This file includes the routines to handle the device part of a send
   for Chameleon

   As a reminder, the first element is the device handle, the second is
   the (basically opaque) mpi handle
 */

/*
   This implementation immediately sends the data.

   It takes advantage of being provided with the address of the user-buffer
   in the contiguous case.
 */
int MPID_CH_post_send( mpid_send_handle, dmpi_send_handle ) 
MPID_SHANDLE *mpid_send_handle;
MPIR_SHANDLE *dmpi_send_handle;
{
char        *address;
MPID_PACKET pkt;
int         actual_len, len;
int         dest;

/* Get the header */
if (mpid_send_handle->start) {
    actual_len = mpid_send_handle->bytes_as_contig;
    }
else {
    DMPI_get_totallen( dmpi_send_handle, &actual_len );
    }
pkt.len        = len = actual_len;
pkt.context_id = mpid_send_handle->context_id;
pkt.tag        = mpid_send_handle->tag;
dest           = mpid_send_handle->to;

if (mpid_send_handle->start) {
    memcpy( pkt.buffer, mpid_send_handle->start, 
	    MPID_MIN(MPID_PACKET_SIZE, actual_len ) );
    }
else {
    DMPI_get_into_contig( dmpi_send_handle, (void *)(pkt.buffer), 
			  MPID_PACKET_SIZE, &actual_len );
    }
if (len > MPID_PACKET_SIZE) {
    /* Send as packet + message */
    PIbsend( MPID_PT2PT_TAG, &pkt, sizeof(MPID_PACKET), dest, MSG_OTHER );
    len -= MPID_PACKET_SIZE;
    mpid_send_handle->sid = 0;
    do {
	if (mpid_send_handle->start) {
	    address = ((char*)mpid_send_handle->start) + MPID_PACKET_SIZE;
	    }
	else {
	    DMPI_get_from_contig( dmpi_send_handle, (void **)(&address), 
				  -1, &actual_len );
	    }
#ifndef PI_NO_NSEND
	if (mpid_send_handle->is_non_blocking) {
	    if (mpid_send_handle->sid)
		/* Not quite correct.  Assumes that only the sid field
		   is used; this is ok as long as the PI_NO_NSEND 
		   wrappers are used... */
		PInwait( MPID_PT2PT2_TAG(PImytid), address, actual_len, 
			 MSG_OTHER, mpid_send_handle->sid );
	    PInsend( MPID_PT2PT2_TAG(PImytid), address, actual_len, dest, 
		     MSG_OTHER, mpid_send_shandle->sid );
	    }
	else 
#endif
	    {
	    PIbsend( MPID_PT2PT2_TAG(PImytid), address, actual_len, dest, 
		     MSG_OTHER );
	    }
	len -= actual_len;
	} while (len > 0);
#ifndef PI_NO_NSEND
    mpid_send_handle->done = 1;
    DMPI_mark_send_completed( dmpi_send_handle );
#else
    if (mpid_send_handle->is_non_blocking) {
	/* As above, the non-blocking version should be used only when 
	   PI_NO_NSEND not defined */
	mpid_send_handle->buffer = address;
	/* mpid_send_handle->actlen = actual_len */
	}
    else {
	mpid_send_handle->done = YES;
	DMPI_mark_send_completed( dmpi_send_handle );
	}
#endif
    }
else {
    /* Always use a blocking send for short messages.
       (May fail with systems that do not provide adequate
       buffering.  These systems should switch to non-blocking sends)
     */
    PIbsend( MPID_PT2PT_TAG, &pkt, actual_len + MPID_HEADER_LEN, dest,
    	     MSG_OTHER );
    mpid_send_handle->done = YES;
    DMPI_mark_send_completed( dmpi_send_handle );
    }
return MPI_OK;
}


/*
  Chameleon gets no asynchronous notice that the message has been complete,
  so there is no asynchronous ref to DMPI_mark_send_completed.
 */
int MPID_CH_isend_wait( mpid_send_handle, dmpi_send_handle )
MPID_SHANDLE *mpid_send_handle;
MPIR_SHANDLE *dmpi_send_handle;
{
/* Wait on the message */
#ifndef PI_NO_NSEND
PIwsend( MPID_PT2PTTAG(mpid_send_handle->to), mpid_send_handle->buffer, 
	 mpid_send_handle->len, mpid_send_handle->to, MSG_OTHER, 
	 mpid_send_handle->sid );
#endif
mpid_send_handle->done = YES;
DMPI_mark_send_completed( dmpi_send_handle );

return MPI_OK;
}



