/* mpir implementation of short list of mpi routines on which the rest can
   be built
*/

#include "mpi.h"
#include "sbcnst.h"

/* Icky global variables */
MPI_COMMUNICATOR MPI_COMM_WORLD;

MPI_Init(argc,argv)
int  argc;
char **argv;
{
    int size;

    /* initialize memory allocation data structures */
    MPIR_shandles = SBinit( sizeof( MPIR_SHANDLE ), 100, 100 );
    MPIR_rhandles = SBinit( sizeof( MPIR_RHANDLE ), 100, 100 );
    MPIR_dtes     = SBinit( sizeof( MPIR_DATATYPE ), 100, 100 );
    MPIR_qels     = SBinit( sizeof( MPIR_QEL ), 100, 100 );
    MPIR_fdtels   = SBinit( sizeof( MPIR_FDTEL ), 100, 100 );

    /* set up pre-defined data types */
    MPIR_int_dte              = (MPIR_DATATYPE *) SBalloc( MPIR_dtes );
    MPIR_int_dte->dte_type    = MPIR_INT;
    MPIR_int_dte->extent      = sizeof( int );
    MPIR_float_dte            = (MPIR_DATATYPE *) SBalloc( MPIR_dtes );
    MPIR_float_dte->dte_type  = MPIR_FLOAT;
    MPIR_float_dte->extent      = sizeof( float );
    MPIR_double_dte           = (MPIR_DATATYPE *) SBalloc( MPIR_dtes );
    MPIR_double_dte->dte_type = MPIR_DOUBLE;
    MPIR_double_dte->extent      = sizeof( double );
    MPIR_long_dte             = (MPIR_DATATYPE *) SBalloc( MPIR_dtes );
    MPIR_long_dte->dte_type   = MPIR_LONG;
    MPIR_long_dte->extent      = sizeof( long );
    MPIR_short_dte            = (MPIR_DATATYPE *) SBalloc( MPIR_dtes );
    MPIR_short_dte->dte_type  = MPIR_SHORT;
    MPIR_short_dte->extent      = sizeof( short );
    MPIR_char_dte             = (MPIR_DATATYPE *) SBalloc( MPIR_dtes );
    MPIR_char_dte->extent      = sizeof( char );
    MPIR_char_dte->dte_type   = MPIR_CHAR;
    MPIR_byte_dte             = (MPIR_DATATYPE *) SBalloc( MPIR_dtes );
    MPIR_byte_dte->extent      = sizeof( char );
    MPIR_byte_dte->dte_type   = MPIR_BYTE;

    /* initialize queues */
    posted_sends.first        = posted_sends.last        = NULL;
    posted_sends.maxlen       = posted_sends.currlen     = 0;

    posted_recvs.first        = posted_recvs.last        = NULL;
    posted_recvs.maxlen       = posted_recvs.currlen     = 0; 

    unexpected_recvs.first    = unexpected_recvs.last    = NULL;
    unexpected_recvs.maxlen   = unexpected_recvs.currlen = 0;

    MPI_COMM_WORLD              = NEW(MPIR_COMMUNICATOR);    
    MPI_COMM_WORLD->comm_type   = MPIR_INTRA;
    MPID_mysize( &size );
    MPI_COMM_WORLD->group       = MPIR_CreateGroup( size );
    MPIR_SetToIdentity( MPI_COMM_WORLD->group );

    MPI_COMM_WORLD->context_pt2pt = MPIR_PT2PT_CONTEXT;
    MPI_COMM_WORLD->context_coll  = MPIR_COLL_CONTEXT;

    /* Place-holders for cache */
    MPI_COMM_WORLD->comm_cache    = 0;
    MPI_COMM_WORLD->attribute_cache = 0;

    /* barrier */
}

MPI_Finalize()
{
    printf("Dumping send queue...\n");
    MPIR_dump_queue( &posted_sends );
    printf("Dumping recv queue...\n");
    MPIR_dump_queue( &posted_recvs );

    /* Free allocated space */
    SBfree( MPIR_dtes, MPIR_int_dte );
    SBfree( MPIR_dtes, MPIR_float_dte );
    SBfree( MPIR_dtes, MPIR_double_dte );
    SBfree( MPIR_dtes, MPIR_long_dte );
    SBfree( MPIR_dtes, MPIR_short_dte );
    SBfree( MPIR_dtes, MPIR_char_dte );
    SBfree( MPIR_dtes, MPIR_byte_dte );

    SBdestroy( MPIR_dtes );
    SBdestroy( MPIR_qels );
    SBdestroy( MPIR_fdtels );
    SBdestroy( MPIR_shandles );
    SBdestroy( MPIR_rhandles );

    MPIR_FreeGroup( MPI_COMM_WORLD->group );
    FREE( MPI_COMM_WORLD );

    /* Tell device that we are done */
    MPID_end();

    /* barrier */
}

MPI_Comm_Init( handle, datatype, source_dest, tag, comm, op_mode, persistence) 
MPI_HANDLE     	 *handle;
MPI_DATATYPE 	 datatype;
int          	 source_dest;
int              tag;
MPI_COMMUNICATOR comm;
MPI_MODE         op_mode;
MPI_PERSISTENCE  persistence;
{

    MPIR_HANDLE *handleptr;
    int         found;

    if (( op_mode == MPI_STANDARD ) ||
	( op_mode == MPI_READY )    ||
	( op_mode == MPI_SYNCHRONOUS ))
    {

	*handle                        = (MPI_HANDLE) SBalloc( MPIR_shandles );
	handleptr                       = *handle;
	handleptr->type                 = MPIR_SEND;
	handleptr->shandle.dest         = source_dest;
	handleptr->shandle.tag          = tag;
	handleptr->shandle.comm         = comm;
	handleptr->shandle.mode         = op_mode;
	handleptr->shandle.persistence  = persistence;
	handleptr->shandle.datatype     = datatype;
	handleptr->shandle.completed    = NO;
	MPID_alloc_send_handle(&((handleptr)->shandle.dev_shandle));
    }
    else if ( op_mode == MPI_RECV )
    {
        MPID_check_for_unexpected( source_dest, tag, comm->context_pt2pt, 
				   &found, &handleptr );
/*	MPIR_search_queue( &unexpected_recvs, source_dest, tag, comm, &found,
			   1, &handleptr ); */
	if (!found)
	{
	    *handle                    = (MPI_HANDLE) SBalloc( MPIR_rhandles );
	    handleptr                       = *handle;
	    handleptr->type                 = MPIR_RECV;
	    handleptr->rhandle.source       = source_dest;
	    handleptr->rhandle.tag          = tag;
	    handleptr->rhandle.comm         = comm;
	    handleptr->rhandle.persistence  = persistence;
	    handleptr->rhandle.datatype     = datatype;
	    handleptr->rhandle.completed    = NO;
	    MPID_alloc_recv_handle(&((handleptr)->rhandle.dev_rhandle));
	}
	else
	{
	    handleptr->rhandle.datatype     = datatype;
	    handleptr->rhandle.persistence  = persistence;
	    MPID_complete_recv( &(handleptr->rhandle.dev_rhandle),
			        &(handleptr->rhandle) );
	}
    }
    else
	MPIR_error("bad op_mode in MPI_COMM_INIT");

    return(MPI_OK);
}

MPI_Comm_Free( handle )
MPI_HANDLE handle;
{
    printf("MPI_Comm_Free called\n");
    if (handle->type == MPIR_SEND)
    {
	MPID_free_send_handle(&((handle)->shandle.dev_shandle));
	SBfree( MPIR_shandles, handle );
    }
    else if (handle->type == MPIR_RECV)
    {
	MPID_free_recv_handle(&((handle)->rhandle.dev_rhandle));
	SBfree( MPIR_rhandles, handle );
    }
    else
	MPIR_error( "MPI_Comm_Free:  bad handle type" );
}

MPI_Start( handle )
MPI_HANDLE handle;
{
    if (handle->type == MPIR_SEND)
    {
	MPID_set_send_dest(& handle->shandle.dev_shandle,handle->shandle.dest);
	MPID_set_send_tag( &handle->shandle.dev_shandle,handle->shandle.tag );
	MPID_set_send_context( &handle->shandle.dev_shandle,handle->shandle.comm->context_pt2pt);
	MPID_set_send_start( &handle->shandle.dev_shandle, NULL );
#ifdef WHEN_WE_DO_THIS
	if (MPIR_IS_CONTIGDATAELEMENT(handle->shandle.datatype)) {
	    MPID_set_send_start( &handle->shandle.dev_shandle, ... );
	    ...
	    }
#endif
	MPIR_enqueue( &posted_sends, handle, MPIR_QSHANDLE );
	MPID_post_send( &(handle->shandle.dev_shandle), &(handle->shandle) );
    }
    else if (handle->type == MPIR_RECV)
    {
	MPID_set_recv_source( &handle->rhandle.dev_rhandle,
			      handle->rhandle.source );
	MPID_set_recv_tag( &handle->rhandle.dev_rhandle, handle->rhandle.tag );
	MPID_set_recv_context( &handle->rhandle.dev_rhandle,handle->rhandle.comm->context_pt2pt);
	MPID_set_recv_start( &handle->rhandle.dev_rhandle, NULL );
#ifdef WHEN_WE_DO_THIS
	if (MPIR_IS_CONTIGDATAELEMENT(handle->rhandle.datatype)) {
	    MPID_set_recv_start( &handle->rhandle.dev_rhandle, ... );
	    ...
	    }
#endif
	MPIR_enqueue( &posted_recvs, handle, MPIR_QRHANDLE );
	MPID_post_recv( &(handle->rhandle.dev_rhandle), &(handle->rhandle) );
    }
    else
	MPIR_error("bad handle type in MPI_START");
}

MPI_Waitall(count, handles, statuses )
int        count;
MPI_HANDLE handles[];
MPI_STATUS *(statuses[]);
{
    int i;
    MPIR_BOOL completed;
    
    for (i = 0; i < count; i++)
    {
        /* This should really use complete_send and complete_recv */
 	if ( handles[i]->type == MPIR_SEND )
	    completed = handles[i]->shandle.completed;
	else
	    completed = handles[i]->rhandle.completed;

	while ( completed == NO ) {
	    MPID_check_device();
	    if ( handles[i]->type == MPIR_SEND )
		completed = handles[i]->shandle.completed;
	    else
		completed = handles[i]->rhandle.completed;
	    }

	if ( handles[i]->type == MPIR_RECV )
	{
	    statuses[i]->source = handles[i]->rhandle.source;
	    statuses[i]->tag    = handles[i]->rhandle.tag;
	    statuses[i]->count  = handles[i]->rhandle.count;  
	}
    }
}

MPI_Testany( count, handles, index, status )
int        count;
MPI_HANDLE handles[];
int        *index;
MPI_STATUS *status;
{
    int i, found;

    found = 0;
    for (i = 0; i < count; i++)
    {
	if ( handles[i] != NULL )
	if ( handles[i]->type == MPIR_SEND )
	    if (handles[i]->shandle.completed == YES)
	    {
		found         = 1;
		*index        = i;
		MPI_Comm_Free( handles[i] ); /* Spec requires, but... */
		handles[i]    = NULL;
		break;
	    }
	else
	    if (handles[i]->rhandle.completed == NO)
	    {
		found         = 1;
		*index        = i;
		status->source = handles[i]->rhandle.source;
		status->tag    = handles[i]->rhandle.tag;
		status->count  = handles[i]->rhandle.count;
		MPI_Comm_Free( handles[i] ); /* Spec requires, but... */
		handles[i]    = NULL;
		break;
	    }
    }
    if ( !found )
	*index = -1;
}

MPI_Probe()
{
#ifdef FOO
    MPIR_search_queue( &unexpected_recvs, source_dest, tag, comm, &found,
			   0, &handleptr );
#endif
}

MPI_Cancel()
{
}

MPI_Comm_Rank( comm, rank )
MPI_COMMUNICATOR comm;
int              *rank;
{
    MPID_myrank(rank);
}

MPI_Address( location, address)
void *location;
int  *address;
{
    *address = (int) location - MPI_BOTTOM;
}

MPI_Extent( datatype, extent )
MPI_DATATYPE datatype;
int          *extent;
{
    *extent = datatype->extent;
}


MPI_Type_Contiguous( count, oldtype, newtype )
int          count;
MPI_DATATYPE oldtype;
MPI_DATATYPE *newtype;
{
    MPIR_DATATYPE  *dteptr;

    *newtype          = (MPI_DATATYPE) SBalloc( MPIR_dtes );
    dteptr            = *newtype;
    dteptr->dte_type  = MPIR_CONTIG;
    dteptr->committed = NO;
    dteptr->extent    = count * oldtype->extent;
    dteptr->use_count = 0;
    dteptr->old_type  = oldtype;
    dteptr->count     = count;
}

MPI_Type_Vector( count, oldtype, stride, blocklen, newtype )
int          count;
MPI_DATATYPE oldtype;
int 	     stride;
int 	     blocklen;
MPI_DATATYPE *newtype;
{
    MPIR_DATATYPE  *dteptr;

    *newtype          = (MPI_DATATYPE) SBalloc( MPIR_dtes );
    dteptr            = *newtype;
    dteptr->dte_type  = MPIR_VECTOR;
    dteptr->committed = NO;
    dteptr->use_count = 0;
    dteptr->old_type  = oldtype;
    dteptr->count     = count;
    dteptr->stride    = stride;
    dteptr->blocklen  = blocklen;
}

MPI_Type_Hvector( count, oldtype, stride, blocklen, newtype )
int          count;
MPI_DATATYPE oldtype;
int 	     stride;
int 	     blocklen;
MPI_DATATYPE *newtype;
{
    MPIR_DATATYPE  *dteptr;

    *newtype          = (MPI_DATATYPE) SBalloc( MPIR_dtes );
    dteptr            = *newtype;
    dteptr->dte_type  = MPIR_HVECTOR;
    dteptr->committed = NO;
    dteptr->use_count = 0;
    dteptr->old_type  = oldtype;
    dteptr->count     = count;
    dteptr->stride    = stride;
    dteptr->blocklen  = blocklen;
}

MPI_Type_Indexed( count, oldtype, indices, blocklens, newtype )
int          count;
MPI_DATATYPE oldtype;
int 	     indices[];
int 	     blocklens[];
MPI_DATATYPE *newtype;
{
    MPIR_DATATYPE  *dteptr;
    int            i;

    *newtype          = (MPI_DATATYPE) SBalloc( MPIR_dtes );
    dteptr            = *newtype;
    dteptr->dte_type  = MPIR_INDEXED;
    dteptr->committed = NO;
    dteptr->use_count = 0;
    dteptr->old_type  = oldtype;
    dteptr->count     = count;
    dteptr->indices   = ( int * ) MPIR_MALLOC( count * sizeof( int ) );
    dteptr->blocklens = ( int * ) MPIR_MALLOC( count * sizeof( int ) );
    for (i = 0; i < count; i++)
    {
	dteptr->indices[i]   = indices[i];
	dteptr->blocklens[i] = blocklens[i];
    }
}

MPI_Type_Hindexed( count, oldtype, indices, blocklens, newtype )
int          count;
MPI_DATATYPE oldtype;
int 	     indices[];
int 	     blocklens[];
MPI_DATATYPE *newtype;
{
    MPIR_DATATYPE  *dteptr;
    int            i;

    *newtype          = (MPI_DATATYPE) SBalloc( MPIR_dtes );
    dteptr            = *newtype;
    dteptr->dte_type  = MPIR_HINDEXED;
    dteptr->committed = NO;
    dteptr->use_count = 0;
    dteptr->old_type  = oldtype;
    dteptr->count     = count;
    dteptr->indices   = ( int * ) MPIR_MALLOC( count * sizeof( int ) );
    dteptr->blocklens = ( int * ) MPIR_MALLOC( count * sizeof( int ) );
    for (i = 0; i < count; i++)
    {
	dteptr->indices[i]   = indices[i];
	dteptr->blocklens[i] = blocklens[i];
    }
}

MPI_Type_Struct( count, oldtypes, indices, blocklens, newtype )
int          count;
MPI_DATATYPE oldtypes[];
int 	     indices[];
int 	     blocklens[];
MPI_DATATYPE *newtype;
{
    MPIR_DATATYPE  *dteptr;
    int            i;

    *newtype          = (MPI_DATATYPE) SBalloc( MPIR_dtes );
    dteptr            = *newtype;
    dteptr->dte_type  = MPIR_STRUCT;
    dteptr->committed = NO;
    dteptr->use_count = 0;
    dteptr->count     = count;
    dteptr->indices   = ( int * ) MPIR_MALLOC( count * sizeof( int ) );
    dteptr->blocklens = ( int * ) MPIR_MALLOC( count * sizeof( int ) );
    dteptr->old_types = ( MPIR_DATATYPE ** )
	                        MPIR_MALLOC( count * sizeof( MPIR_DATATYPE *));
    for (i = 0; i < count; i++)
    {
	dteptr->indices[i]   = indices[i];
	dteptr->blocklens[i] = blocklens[i];
	dteptr->old_types[i] = oldtypes[i];
    }
}

